from cnvrg.modules.base_module import CnvrgBase
import cnvrg.helpers.apis_helper as apis_helper
from cnvrg.modules.errors import CnvrgError
from cnvrg.helpers.config_helper import config_type, CONFIG_TYPE_PROJECT
from cnvrg.modules.project import Project
import os
import re
class Library(CnvrgBase):
    def __init__(self, library):
        self.__cloned = False
        self.__library = library
        self.__owner = apis_helper.credentials.owner
        self.__info = None
        self.__path = None

    def __base_url(self):
        return os.path.join("users", self.__owner, "libraries", self.__library)


    @staticmethod
    def list():
        owner = apis_helper.credentials.owner
        return apis_helper.get(os.path.join("users", owner, "libraries")).get("libraries")


    def info(self, force=False):
        """
        get info about this library
        :param force: to force api fetching
        :return: dict represent the library
        {clone_cmd: string,
        command: string,
        arguments: list of key: values(list),
        number_of_experiments: integer,
        description: string}
        """
        if self.__info and not force:
            return self.__info
        self.__info = self.__fetch_info()
        return self.info()


    def __fetch_info(self):
        resp = apis_helper.get(self.__base_url())
        return resp.get("library")



    def load(self, working_dir=None):
        """
        load library to your local directory
        :param working_dir: path to clone the library to
        """
        working_dir = working_dir or os.curdir
        if self.__lib_loaded(working_dir=working_dir):
            self.__path = self.__lib_path(working_dir=working_dir)
            return self.__path
        info = self.info()
        lib_dir = self.__lib_path(working_dir=working_dir)
        cmds = [
            "mkdir -p {working_dir}".format(working_dir=lib_dir),
            "cd {working_dir}".format(working_dir=lib_dir),
            info.get("clone_cmd"),
        ]
        if config_type(working_dir) == CONFIG_TYPE_PROJECT:
            ## if inside project, add the file to the gitignore
            cmds.append("echo 'libraries/**' >> .cnvrgignore")

        cmds = " && ".join(cmds)

        #run the commands
        returncode = os.system(cmds)
        if returncode != 0:
            raise CnvrgError("Cant clone library")
        self.__path = lib_dir
        return lib_dir

    def __add_args(self, args=None):
        command = self.info().get("command")
        args = args or {}
        library_args = self.info().get("arguments") or []
        merged_args = {x["key"]:x["value"] for x in library_args}
        merged_args = {**merged_args, **args}
        string_args = " ".join(map(lambda x: "--{key}={value}".format(key=x[0], value=x[1]), merged_args.items()))
        return "{cmd} {args}".format(cmd=command, args=string_args)

    def run(self, templates=None, datasets=None, image=None, commit=None, title=None, project_url=None, local=True, args=None):
        if local: return self.run_local(args=args)
        project = Project(project_url=project_url)
        exp = project.run_experiment(cmd=self.info().get("command"), templates=templates, datasets=datasets, image=image, commit=commit, title=title)
        return exp




    def run_local(self, working_dir=None, args=None):
        working_dir = working_dir or self.__path
        if config_type(os.curdir) != CONFIG_TYPE_PROJECT:
            raise CnvrgError("Cant run a library not in a project context, please cd into cnvrg project")
        raw_script = self.__add_args(args)
        script = re.sub(self.info().get("file"), os.path.join(os.path.realpath(working_dir), self.info().get("file")), raw_script)
        commands = " && ".join([
            "cnvrg run -l {script}".format(script=script)
        ])
        os.system(commands)

    def __lib_loaded(self, working_dir=None):
        return os.path.exists(os.path.join(self.__lib_path(working_dir=working_dir), ".cnvrg", "config.yml"))


    def __lib_path(self, working_dir=None):
        working_dir = working_dir or os.curdir
        return os.path.join(working_dir, "libraries", self.info().get("title"))











from cnvrg.modules.cnvrg_job import CnvrgJob, LOGS_TYPE_OUTPUT, LOGS_TYPE_ERROR
import cnvrg.helpers.param_build_helper as param_build_helper
from cnvrg.modules.project import Project
import cnvrg.helpers.logger_helper as logger_helper
import cnvrg.helpers.string_helper as string_helper
from cnvrg.modules.errors import UserError
import cnvrg.helpers.apis_helper as apis_helper
import cnvrg.helpers.env_helper as env_helper
import cnvrg.helpers.chart_show_helper as chart_show_helper
from cnvrg.helpers.env_helper import in_experiment, CURRENT_JOB_ID
from enum import Enum
from typing import List
import os
import time


class TagType(Enum):
    SINGLE_TAG = "single"
    LINECHART_TAG = "linechart"



class Experiment(CnvrgJob):
    def __init__(self, experiment):
        owner, project_slug, slug = param_build_helper.parse_params(experiment, param_build_helper.EXPERIMENT)
        if not in_experiment() and not slug:
            raise UserError("Cant create an experiment without slug and outside experiment run")
        slug = slug or CURRENT_JOB_ID
        super(Experiment, self).__init__(slug, env_helper.EXPERIMENT, Project(project_slug))
        self.__data = self.__get_experiment()



    def start(self):
        pass

    def log_output(self):
        pass

    def tag(self, key, value=None):
        tag_data = {
            "key": key,
            "value": value,
            "type": TagType.SINGLE_TAG.value
        }
        self.__send_tag(tag_data)

    def chart(self, key, Ys: List, Xs: List=None, grouping: List=None, x_axis=None, y_axis=None) -> None:
        """
        a function which can tag an experiment with a chart
        :param key: the name of the chart
        :param Ys: [y1, y2, y3, y4] (float)
        :param Xs: [x1, x2, x3, x4] (date, integer, null)
        :param grouping: [g1, g2, g3, g4]
        :param x_axis: rename the x_axis of the chart
        :param y_axis:rename the y_axis of the chart
        :return:
        """
        tag_data = {
            "ys": Ys,
            "xs": Xs,
            "key": key,
            "grouping": grouping,
            "x_axis": x_axis,
            "y_axis": y_axis,
            "type": TagType.LINECHART_TAG.value,
        }
        self.__send_tag(tag_data)

    def logs(self, callback=None, poll_every=5):
        job_logs, experiment_is_running = self.__fetch_logs(0)
        offset = len(job_logs)
        callback = callback or logger_helper.log_cnvrg_log
        [callback(l) for l in job_logs]
        while experiment_is_running:
            time.sleep(poll_every)
            job_logs, experiment_is_running = self.__fetch_logs(offset)
            offset += len(job_logs)
            [callback(l) for l in job_logs]


    def show_chart(self, key, **kwargs):
        """

        :param key: chart_key
        :param kwargs: with_legend, legend_loc
        :return:
        """
        chart = apis_helper.get(os.path.join(self._base_url(), 'charts', key)).get("chart")
        return chart_show_helper.show_chart(chart, **kwargs)


    def __fetch_logs(self, offset, limit=None):
        resp = apis_helper.get(os.path.join(self._base_url(), 'logs'), data={"offset": offset, "limit": limit})
        return resp.get("logs"), resp.get("experiment").get("is_running")


    def __send_tag(self,tag_data):
        apis_helper.post(os.path.join(self._base_url(), 'tags'), data={"tag": tag_data})

    def finish(self, exit_status=None):
        pass

    def __get_experiment(self):
        return apis_helper.get(self._base_url()).get("experiment")

    def _base_url(self):
        return os.path.join(
            #### hackish :D
            self.project.get_base_url(),string_helper.to_snake_case(self.job_type) + "s", self.job_slug
        )

    def __getitem__(self, item):
        return self.__data.get(item)






from cnvrg.modules.experiment import Experiment
from cnvrg.modules.experiment import LOGS_TYPE_OUTPUT, LOGS_TYPE_ERROR
import cnvrg.helpers.spawn_helper as spawn_helper
from cnvrg.libs.limit import limit
import queue
import time
from threading import Thread
import io


STOP_MESSAGE = "@AFSDFAQ#$TASDFQ@#$RKASDLFAKSDLFM<AQL@#$MTLAS<DLF<AWLMNOQ#K$%LAW<EFL:A<>D:ALS:FW$<TRLEMFLAMSDFAsf"
SLEEP_DELAY = 0.001


def __logging_thread(experiment, pid, error_log=False):
    log_type = LOGS_TYPE_ERROR if error_log else LOGS_TYPE_OUTPUT
    q = queue.Queue()
    fid = pid.stderr if error_log else pid.stdout
    t = Thread(target=log_exp, args=(experiment, q, log_type))
    t.start()
    time.sleep(1)
    print("start")
    for line in iter(fid.readline, b''):
        print(line.rstrip())
        q.put(line.decode("utf-8").rstrip())
    print("IM HERE!!")
    fid.close()
    q.put(STOP_MESSAGE)
    t.join()

def __monitoring_thread(experiment, pid):
    ### check every second if the process died, send utilization every 15 seconds.
    while pid.poll() != None:
        for i in range(30):
            time.sleep(1)
            if pid.poll() != None:
                return
        monitor(experiment, pid)

def monitor(experiment, pid):
    if pid.poll() != None:
        return
    utilization = spawn_helper.analyze_pid(pid)
    experiment.send_util(utilization)

def log_exp(experiment: Experiment, q: queue.Queue, log_type="output"):
    while True:
        if q.empty():
            print("Q is empty")
            time.sleep(0.5)
            continue
        logs = []
        while not q.empty():
            logs.append(q.get())
        print(logs)
        experiment.log(list(filter(lambda x: x is not STOP_MESSAGE, logs)), log_type=log_type)
        if STOP_MESSAGE in logs:
            return
        time.sleep(3)

def exec(exp_slug):
    experiment = Experiment(exp_slug)
    cmd = experiment["input"]
    pid = spawn_helper.__send_cmd(cmd)
    logger = Thread(target=__logging_thread, args=(experiment, pid, False))
    logger.start()
    #error_logger = Thread(target=__logging_thread, args=(experiment, pid, True))
    #error_logger.start()
    utilization = Thread(target=__monitoring_thread, args=(experiment, pid))
    utilization.start()
    while pid.poll() == None:
        time.sleep(0.5)
    utilization.join()
    #error_logger.join()
    logger.join()

    experiment.finish(exit_status=pid.poll())






import os

EXPERIMENT = "Experiment"


POOL_SIZE = os.environ.get("CNVRG_THREAD_SIZE") or 100
CURRENT_JOB_ID = os.environ.get("CNVRG_JOB_ID")
CURRENT_JOB_TYPE = os.environ.get("CNVRG_JOB_TYPE")
MAX_LOGS_PER_SEND = int(os.environ.get("CNVRG_MAX_LOGS_PER_SEND") or 500)

def in_experiment():
    return CURRENT_JOB_TYPE == EXPERIMENT


import click
import cnvrg.helpers.cli_helper as cli_helper
from cnvrg.modules.library import Library
from cnvrg.helpers.logger_helper import log_message
import os

@click.group()
def library():
    pass

@library.group("library")
def lib_cmd():
    pass

@lib_cmd.command()
@click.argument("library_title")
def info(library_title):
    l = Library(library_title)
    cli_helper.print_object(l.info())


@lib_cmd.command()
def list():
    cli_helper.print_object(Library.list())


@lib_cmd.command()
@click.argument("library_title")
@click.option('--working_dir', '-w', default=os.getcwd(), help='working dir')
def clone(library_title, working_dir):
    l = Library(library_title)
    library_path = l.load(working_dir=working_dir)
    log_message("You can find the library in {path}".format(path=library_path), fg='green')



@lib_cmd.command(help="Title")
@click.argument("library_title")
@click.option("--local/--remote", default=False)
@click.option("--arg", multiple=True)
@click.option("--dataset", multiple=True)
@click.option("--compute", multiple=True)
@click.option("--working_dir", help="directory to run from")
@click.option("--title", help="specify experiment title")
def run(library_title, local, arg, dataset, working_dir, compute, title):
    arguments = cli_helper.parse_args(arg)
    datasets = cli_helper.parse_datasets(dataset)
    l = Library(library_title)
    if local:
        l.load(working_dir=working_dir)
        l.run_local(args=arguments)
        return
    cli_helper.print_object(l.run(datasets=datasets, templates=compute, title=title))


import os
from typing import Dict
from cnvrg.helpers.hash_helper import hash_file
import mimetypes

def get_file_props(file: str, working_dir: str) -> Dict:
    '''

    :param file:
    :param working_dir:
    :return: {relative_path, file_name, file_size, content_type, sha1}
    '''
    abs_file_path = os.path.realpath(os.path.join(working_dir, file))
    relative_path = os.path.relpath(file, working_dir)
    size = os.path.getsize(abs_file_path)
    file_name = os.path.basename(abs_file_path)
    content_type = mimetypes.MimeTypes().guess_type(abs_file_path)[0] or "text/plain"
    return {
        "relative_path": relative_path,
        "file_name": file_name,
        "file_size": size,
        "content_type": content_type,
        "sha1": hash_file(abs_file_path, raw=True)
    }


from cnvrg.modules.storage.storage import Storage
from cnvrg.modules.storage.s3_storage import S3Storage
from cnvrg.modules.project import Project
from cnvrg.helpers.apis_helper import post as apis_post, get as apis_get, download_file
import os


STORAGE_TYPE_GCP = "gcp"
STORAGE_TYPE_S3 = "s3"
STORAGE_TYPE_MINIO = "minio"

def storage_factory(element: Project, working_dir=os.curdir):
    storage_resp = apis_get(element.get_storage_url()).get("client")
    if storage_resp.get("storage") == STORAGE_TYPE_S3:
        return S3Storage(element, working_dir, storage_resp)


import os
import re
from pathlib import Path
def ignore_path(path):
    return os.path.join(path,'.cnvrgignore')

def generate_cnvrgignore(path):
    content = """# cnvrg ignore: Ignore the following directories and files
    # for example:
    # some_dir/
    # some_file.txt
    .git/*
    .cnvrg/*
    .gitignore
    *.conflict
    *.deleted"""
    with open(ignore_path(path), 'w') as f:
        f.write(content)

def check_cnvrgignore(path):
    if not os.path.exists(ignore_path(path)):
        generate_cnvrgignore(path)

def blank_ignores(f):
    return not f.startswith("#")

def remove_blanks(f):
    return re.sub("\s", "", f)

def read_cnvrgignore(path):
    i_path = ignore_path(path)
    if not os.path.exists(i_path): return []
    with open(i_path, 'r') as f:
        return (filter(blank_ignores, map(remove_blanks, f.readlines())))

def get_full_ignore(path):
    ignores = []
    for ignore in (list(read_cnvrgignore(path)) + [".cnvrg/*"]):
        if ignore.endswith("/"): ignore = ignore + "**/*"
        elif os.path.isdir(ignore): ignore = ignore + "/**/*"
        elif not ignore: continue
        files = list(Path(path).glob(ignore))
        ignores.extend(files)
    return list(map(lambda x: str(x), ignores))


import yaml
import re
import os
import requests
import cnvrg.modules.errors as errors
from tinynetrc import Netrc
from pathlib import Path
NETRC_HOST = "cnvrg.io"
CONFIG_FILE_PATH = os.path.join(os.path.expanduser("~"), ".cnvrg", "config.yml")
NETRC_FILE_PATH = os.path.join(os.path.expanduser("~"), '.netrc')


class CnvrgCredentials():

    def __init__(self):
        self.token = None
        self.api_url = self.set_api_url("https://app.cnvrg.io/api")
        self.owner = None
        self.username = None
        self.logged_in = self._load_yaml() or self._load_environ()


    def set_api_url(self, api_url):
        api_url = re.sub(r'(\/api\/?)?(v1.*)?', '', api_url)
        self.api_url = os.path.join(api_url, 'api')
        return self.api_url

    def login(self, email, password, api_url=None):
        api_url = self.set_api_url(api_url or self.api_url)
        resp = requests.post(os.path.join(api_url, 'v1', 'users', 'sign_in'), headers={"EMAIL": email, "PASSWORD": password})
        if resp.status_code != 200:
            raise errors.CnvrgError("Can't Authenticate {email}".format(email=email))

        res = resp.json().get("result")
        token = res.get("token")
        username = res.get("username")
        owner = res.get("owners")[0]
        api_url = api_url or res.get("urls")[0]
        self.__set_credentials(token=token, owner=owner, username=username, email=email, api_url=api_url)
        self.logged_in = True
        self.username = username

    def logout(self):
        if not self.logged_in: return
        netrc= Netrc()
        del netrc[NETRC_HOST]
        netrc.save()
        os.remove(CONFIG_FILE_PATH)
        return True


    def __set_credentials(self, token=None, owner=None, username=None, email=None, api_url=None):
        Path(NETRC_FILE_PATH).touch()
        netrc = Netrc()
        netrc[NETRC_HOST] = {"login": email, "password": token}
        netrc.save()

        os.makedirs(os.path.dirname(CONFIG_FILE_PATH), exist_ok=True)
        Path(CONFIG_FILE_PATH).touch()
        with open(CONFIG_FILE_PATH, "w") as cnvrg_config:
            yaml.dump({":owner": owner, ":username": username, ":api": api_url}, cnvrg_config)





    def _load_environ(self):
        token = os.environ.get("CNVRG_AUTH_TOKEN")
        api_url = os.environ.get("CNVRG_API")
        owner = os.environ.get("CNVRG_OWNER")
        if not api_url: return None
        self.set_api_url(api_url)
        if not token: return None
        if not owner: return None
        self.token = token
        self.owner = owner
        return True

    def _load_yaml(self):
        if not os.path.exists(NETRC_FILE_PATH): return None
        if not os.path.exists(CONFIG_FILE_PATH): return None
        netrc = Netrc()
        token = netrc[NETRC_HOST]["password"]
        config = yaml.safe_load(open(CONFIG_FILE_PATH, "r"))
        api_url = config.get(":api") or config.get("api")
        owner = config.get(":owner") or config.get("owner")
        if not api_url: return None
        self.set_api_url(api_url)
        if not token: return None
        if not owner: return None
        self.token = token
        self.owner = owner
        return True


import requests
import json
import os
from cnvrg.modules.errors import CnvrgError
import cnvrg.helpers.auth_helper as auth_helper
import cnvrg.helpers.logger_helper as logger_helper


JSON_HEADERS = {
    "Content-Type": "application/json"
}
def verify_logged_in():
    if not credentials.logged_in:
        raise CnvrgError("Not authenticated")

def __parse_resp(resp, **kwargs):
    try:
        r_j = resp.json()
        return r_j
    except Exception as e:
        logger_helper.log_error(e)
        logger_helper.log_bad_response(**kwargs)



def get(url, data=None):
    verify_logged_in()
    get_url = os.path.join(base_url, url)
    resp = session.get(get_url, params=data)
    return __parse_resp(resp, url=url, data=data)

def post(url, data=None):
    verify_logged_in()
    get_url = os.path.join(base_url, url)
    resp = session.post(get_url, data=json.dumps(data))
    return __parse_resp(resp, url=url, data=data)

def put(url, data=None):
    verify_logged_in()
    get_url = os.path.join(base_url, url)
    resp = session.put(get_url, data=json.dumps(data))
    return __parse_resp(resp, url=url, data=data)

def download_file(url):
    verify_logged_in()
    resp = requests.get(url)
    return resp.text

credentials = auth_helper.CnvrgCredentials()
session = requests.session()
session.headers = {
    "AUTH_TOKEN": credentials.token,
    **JSON_HEADERS
}
base_url = os.path.join(credentials.api_url, "v1")




from cnvrg.modules.storage import Storage
from cnvrg.modules.project import Project
from cnvrg.modules.storage import storage_factory
from cnvrg.helpers.status_helpers import CommitStatus
from cnvrg.actions import check_if_path_exists
import os
from cnvrg.helpers import log_message
from cnvrg.modules.commituploader import CommitUploader


def clone(project_url, commit_sha1=None, working_dir=None):
    ## creating project
    project = Project(project_url=project_url, working_dir=working_dir)
    log_message("Cloning {project}".format(project=project.get_project_name()))
    project_path = os.path.join(working_dir or os.curdir, project.get_project_name())
    project._set_working_dir(project_path)

    ## handle already exists
    check_if_path_exists(project_path)
    os.makedirs(project_path, exist_ok=True)

    #init storage
    storage = storage_factory(project, working_dir=project_path)
    commit = project.fetch_commit(commit_sha1 or "latest")

    #download
    storage.download_dirs(commit.get("trees"))
    storage.download_files(commit.get("blobs"))

    #post download
    project._clone_finish(commit.get("sha1"), project_path)

    log_message("{project} cloned successfuly".format(project=project.get_project_name()))


def download(storage: Storage=None, commit_sha1=None, force=False):
    storage = storage or storage_factory(Project())
    element = storage.element
    status_o = element.status(to_commit=commit_sha1)
    files_to_download = status_o[CommitStatus.UPDATED_ON_SERVER]
    files_to_delete = status_o[CommitStatus.DELETED_ON_REMOTE]

    ####
    # we need to rename all the delete with conflicts to FILENAME.deleted
    # we need to download all the files from updated on server with the name of FILENAME.conflict
    ####
    files_to_delete_with_conflict = status_o[CommitStatus.DELETED_CONFLICTS] if not force else []
    conflicts = status_o[CommitStatus.CONFLICTS] if not force else []
    ####

    ### Latest commit or commit to download
    commit_sha1 = status_o[CommitStatus.COMMIT_SHA1]
    if commit_sha1 == element.get_current_commit():
        log_message("Already up to date")
        return

    if len(files_to_download) > 0:
        files = storage.element.get_commit_files(commit_sha1=commit_sha1, files=files_to_download)
        log_message("Downloading {count} files".format(count=len(files)))
        storage.download_files(files, conflicts=conflicts)
    if len(files_to_delete) > 0:
        storage.delete_files(files_to_delete, conflicts=files_to_delete_with_conflict)
    storage.element._download_finish(commit_sha1)
    log_message("Project downloaded successfuly")


def upload(storage: Storage=None, force=False, new_branch=False):
    storage = storage or storage_factory(Project())
    status_o = storage.element.status(force=force, new_branch=new_branch)
    CommitUploader(storage, status_o)





import json
def parse_args(args):
    return dict(map(lambda x: x.split(":"), args))


def parse_datasets(datasets):
    parsed_datasets = []
    for dataset in datasets:
        commit = None
        dataset_slug = dataset
        if ":" in dataset:
            dataset_slug, commit = dataset.split(":")
        parsed_datasets += [{"slug": dataset_slug, "commit": commit}]
    return parsed_datasets



def print_object(o):
    print(json.dumps(o, sort_keys=True, indent=4))